import asyncio
import os
import random
import urllib.parse

import aiofiles
from fastapi import Depends, APIRouter, HTTPException
from starlette.requests import Request

from .configs import config
from .file_shard import FileShard
from .lifespan import TridentFastAPIRequest

router = APIRouter()


async def parse_body(request: Request):
    data: bytes = await request.body()
    return data


@router.put("/kv/{key}")
async def store(
    request: TridentFastAPIRequest,
    key: str,
    value: bytes = Depends(parse_body),
    dry_run: bool = False,
):
    writing_file_shards = []

    for point in request.app.state.hash_ring.range(
        key, config["replicas"].get(int), unique=True
    ):
        file_shard: FileShard = point["instance"]
        writing_file_shards.append(file_shard)

    if not dry_run:
        await asyncio.gather(
            *[
                writing_file_shard.write_file(key, value)
                for writing_file_shard in writing_file_shards
            ]
        )

    return {
        "file_shards": [
            writing_file_shard.name for writing_file_shard in writing_file_shards
        ]
    }


@router.get("/kv/{key}")
async def read(request: TridentFastAPIRequest, key: str) -> bytes:
    points = list(
        request.app.state.hash_ring.range(key, config["replicas"].get(int), unique=True)
    )
    random.shuffle(points)
    for point in points:
        file_shard: FileShard = point["instance"]
        file = await file_shard.read_file(key)
        if file:
            return file
    raise HTTPException(status_code=404, detail="not_found")



import dataclasses
from contextlib import asynccontextmanager

from fastapi import FastAPI
from izihawa_loglib import configure_logging
from starlette.requests import Request
from uhashring import HashRing

from .configs import config
from .file_shard import FileShard


@dataclasses.dataclass
class State:
    hash_ring: HashRing


class TridentFastAPI(FastAPI):
    state: State


class TridentFastAPIRequest(Request):
    app: TridentFastAPI


@asynccontextmanager
async def lifespan(app: TridentFastAPI):
    configure_logging(config.get(dict))
    nodes = {
        node_config["name"]: {
            "instance": FileShard(node_config["name"], node_config["path"]),
            "weight": node_config["weight"],
        }
        for node_config in config["file_shards"].get(list)
    }
    app.state = State(hash_ring=HashRing(nodes))
    yield



import os
import urllib.parse

import aiofiles


class FileShard:
    def __init__(self, name: str, path: str):
        self._name = name
        self._path = path

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return self._path

    async def read_file(self, key: str) -> bytes:
        file_path = os.path.join(self.path, urllib.parse.quote(key))
        if os.path.exists(file_path):
            async with aiofiles.open(file_path, 'rb') as f:
                return await f.read()

    async def write_file(self, key: str, value: bytes):
        file_path = os.path.join(self.path, urllib.parse.quote(key))
        async with aiofiles.open(file_path, "wb") as f:
            await f.write(value)



from aiobaseclient import BaseClient
from aiobaseclient.exceptions import ExternalServiceError


class TridentClient(BaseClient):
    async def response_processor(self, response):
        data = await response.read()
        if response.status == 404:
            return None
        elif response.status != 200:
            if hasattr(response, 'request'):
                raise ExternalServiceError(response.request.url, response.status, data)
            else:
                raise ExternalServiceError(None, response.status, data)
        return data

    async def store(self, key: str, data: bytes):
        url = f'/kv/{key}'
        return await self.put(url, data=data)

    async def read(self, key: str) -> bytes:
        url = f'/kv/{key}'
        return await self.get(url)



import logging
import urllib.parse
import zipfile

import fire
from aiosumma import SummaClient
from trident.client import TridentClient



def process_zip(filepath):
    with zipfile.ZipFile(filepath) as archive:
        zip_infos = archive.infolist()
        # iterate through each file
        for zip_info in zip_infos:
            prefix, suffix = zip_info.filename.lower().split('/', 1)
            suffix = urllib.parse.unquote(suffix)
            doi = prefix + '/' + suffix
            data = archive.read(zip_info)
            yield [doi, data]


async def local_import(summa_client, trident_client, local_path: str):
    for doi, data in process_zip(local_path):
        document = await summa_client.get_one_by_field_value(
            "nexus_science",
            "id.dois",
            doi,
        )
        if not document:
            logging.getLogger('statbox').info({
                'action': 'not_found',
                'doi': doi,
            })
            return

        if 'nexus_id' not in document:
            continue

        key = f'/kv/{document["id"]["nexus_id"]}.pdf'
        await trident_client.store(key, data)


async def import_file(summa_endpoint, trident_base_url, local_path):
    summa_client = SummaClient(endpoint=summa_endpoint.get(str))
    trident_client = TridentClient(base_url=trident_base_url.get(str))
    await summa_client.start()
    await trident_client.start()

    await local_import(summa_client, trident_client, local_path)


def main():
    fire.Fire(import_file)
